/* $LAAS: log.cc 2008/12/08 15:05:29 mallet $ */

/* 
 * Copyright (c) 2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Sat Aug  2 2008
 */

static const char _rcsid[] = "$LAAS: log.cc 2008/12/08 15:05:29 mallet $";

#include <err.h>
#include <string.h>
#include <stdlib.h>

#include "codels.h"


/* --- hrp2_logreadstate --------------------------------------------- */

/** \brief Read robot state from log
 *
 * Records must be separated by ^L
 */

int
hrp2_logreadstate(FILE *log, OpenHRP::RobotState_out state)
{
  char line[1024], *token, *next;
  unsigned int i;

  /* create a new robot state */
  state = new OpenHRP::RobotState;
  state->angle.length(GHRP2_MAXDOF);
  state->attitude.length(1);
  state->attitude[0].length(9);
  state->force.length(4);
  for(i=0; i<4; i++) state->force[i].length(6);
  state->rateGyro.length(1);
  state->rateGyro[0].length(3);
  state->accel.length(1);
  state->accel[0].length(3);

  /* parse a record */
  while(fgets(line, sizeof(line), log)) {
    token = strtok_r(line, " \t", &next);

    /* end-of-record */
    if (!strcmp(token, "\f\n")) {
      /* XXX this is temporary because there is no log (yet) for
       * attitude: we have to recompute the value in the chest frame */
      T3D c2w, w2w;

      t3dInit(&c2w, T3D_BRYAN, T3D_ALLOW_CONVERSION);
      c2w.bryan.yaw = state->angle[GHRP2_CHEST_JOINT0];
      c2w.bryan.pitch = state->angle[GHRP2_CHEST_JOINT1];

      t3dInit(&w2w, T3D_BRYAN, T3D_ALLOW_CONVERSION);
      w2w.bryan.yaw = state->attitude[0][0];
      w2w.bryan.pitch = state->attitude[0][1];
      w2w.bryan.roll = state->attitude[0][2];
      t3dComp(&c2w, &w2w);

      t3dConvertTo(T3D_BRYAN, &c2w);
      state->attitude[0][0] = c2w.bryan.roll;
      state->attitude[0][1] = c2w.bryan.pitch;
      state->attitude[0][2] = c2w.bryan.yaw;

      return 0;
    }

#define parse_record_entry(k, s, d)			\
    do if (!strcmp(token, k)) {				\
      for(i=0, token = strtok_r(NULL, " \t", &next);	\
	  i<(s) && token;				\
	  i++, token = strtok_r(NULL, " \t", &next))	\
	(d)[i] = atof(token);				\
    } while(0)

    /* fill state with record entries */
    parse_record_entry
      ("config:", GHRP2_MAXDOF, state->angle);
    parse_record_entry
      ("right-leg-force:", 6, state->force[0]);
    parse_record_entry
      ("left-leg-force:", 6, state->force[1]);
    parse_record_entry
      ("rpy-imu:", 9, state->attitude[0]);
    parse_record_entry
      ("imu:", 6, (i<3)?&state->rateGyro[0][0]:(&state->accel[0][0]-3));

#undef parse_record_entry
  }

  /* eof, or error */
  if (ferror(log)) warn("while reading playback log");
  return -1;
}


/* --- hrp2_logwritestate -------------------------------------------- */

/** \brief Write robot state to log
 *
 */

int
hrp2_logwritestate(FILE *log, OpenHRP::RobotState_var state)
{
  hrp2_logwritedrecord(log, "config", GHRP2_MAXDOF, &state->angle[0]);
  hrp2_logwritedrecord(log, "imu-rpy", 9, &state->attitude[0][0]);
  hrp2_logwritedrecord(log, "imu-rate", 3, &state->rateGyro[0][0]);
  hrp2_logwritedrecord(log, "imu-accel", 3, &state->accel[0][0]);
  hrp2_logwritedrecord(log, "force-right-leg", 6, &state->force[0][0]);
  hrp2_logwritedrecord(log, "force-left-leg", 6, &state->force[1][0]);
  return 0;
}


/* --- hrp2_logwritedrecord ------------------------------------------ */

/** \brief Write an array of double to log
 *
 */

int
hrp2_logwritedrecord(FILE *log,
		     const char *title, unsigned int elems, const double *array)
{
  unsigned int i;

  fprintf(log, "%s: ", title);
  for (i=0; i<elems; i++) fprintf(log, "%12g ", *(array++));
  fprintf(log, "\n");
  return 0;
}


/* --- hrp2_logwritedrecord ------------------------------------------ */

/** \brief Write end-of-record marker
 *
 */

int
hrp2_logend(FILE *log)
{
  fprintf(log, "\f\n");
  return 0;
}
