/* $LAAS: odometry.c 2008/08/03 23:21:13 tho $ */

/*
 * Copyright (c) 2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Wed Jul 23 2008
 */

static const char _rcsid[] = "$LAAS: odometry.c 2008/08/03 23:21:13 tho $";

#include <math.h>

#include "codels.h"


/* --- local data ---------------------------------------------------- */

#define HRP2_ODORIGHT	(0)
#define HRP2_ODOLEFT	(1)

static T3D Thip[2] = {
  { .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		   [T3D_BRYAN_Y] = 0.060
		 }}}},
  { .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		   [T3D_BRYAN_Y] = -0.060
		 }}}}
};

static T3D Tthigh = {
  .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		 [T3D_BRYAN_Z] = 0.300
	       }}}
};

static T3D Tleg[2] = {
  { .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		   [T3D_BRYAN_Y] = 0.035, [T3D_BRYAN_Z] = 0.300
		 }}}},
  { .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		   [T3D_BRYAN_Y] = -0.035, [T3D_BRYAN_Z] = 0.300
		 }}}}
};


/** left foot position wrt origin */
static T3D lfoot = {
  .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		 [T3D_BRYAN_X] = 0.00949035,
		 [T3D_BRYAN_Y] = 0.095,
		 [T3D_BRYAN_Z] = 0.105,
	       }}}};

/** right foot position wrt origin */
static T3D rfoot = {
  .bryan = { T3D_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK, {{
		 [T3D_BRYAN_X] = 0.00949035,
		 [T3D_BRYAN_Y] = -0.095,
		 [T3D_BRYAN_Z] = 0.105,
	       }}}};

/** left foot bias (roll pitch) at the contact point */
static T3D lfootbias = {
  .bryan = { T3D_U_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK }
};

/** right foot bias (roll pitch) at the contact point */
static T3D rfootbias = {
  .bryan = { T3D_U_BRYAN, T3D_ALLOW_CONVERSION, T3D_EOK }
};


static int	get_waist2foot(const double angles[GHRP2_MAXDOF], int leg,
			T3D *rt3d);
static int	get_footbias(const T3D *w2f, const T3D *f2W, double imupitch,
			double imuroll, double force, T3D *bias);


/* --- hrp2_odosetpos ------------------------------------------------ */

/** \brief Set actual odometry position
 *
 * \param[in] angles	Actual robot configuration
 * \param[in] waist	Actual waist position to be set
 */

int
hrp2_odosetpos(const double angles[GHRP2_MAXDOF], const T3D *waist)
{
  T3D w;
  T3D w2f;

  t3dCopyIn(&w, waist);
  w.flags |= T3D_ALLOW_CONVERSION;

  if (get_waist2foot(angles, HRP2_ODOLEFT, &w2f)) return -1;
  if (!t3dCompIn(&lfoot, t3dInvert(&w2f), &w)) return -1;

  if (get_waist2foot(angles, HRP2_ODORIGHT, &w2f)) return -1;
  if (!t3dCompIn(&rfoot, t3dInvert(&w2f), &w)) return -1;

  return 0;
}


/* --- hrp2_odogetpos ------------------------------------------------ */

/** \brief Update odometry position
 *
 * \param[in] angles	Actual robot configuration.
 * \param[in] lfootF	Force sensor values for left foot.
 * \param[in] lfootR	Force sensor values for right foot.
 * \param[in] imupitch	Pitch of the waist from IMU.
 * \param[in] imuroll	Roll of the waist from IMU.
 * \param[out] waist	Actual waist position.
 */

int
hrp2_odogetpos(const double angles[GHRP2_MAXDOF],
	       const double rfootF[6], const double lfootF[6],
	       double imupitch, double imuroll,
	       T3D *waist)
{
  T3D w2f;

  if ((rfootF[2] >= HRP2_FORCETH && lfootF[2] >= HRP2_FORCETH) ||
      (rfootF[2] < HRP2_FORCETH && lfootF[2] < HRP2_FORCETH)) {
    /* both feet are fixed or both feet are in the air: compute waist
     * from left and right and average the result. Of course this doesn't
     * make much sense in the case where both feet are not fixed, but at
     * least we'll get the pitch and roll correct. */
    T3D avg = {{ .flags = T3D_ALLOW_CONVERSION }};

    /* waist from right foot and correct with IMU */
    unlikely(get_waist2foot(angles, HRP2_ODORIGHT, &w2f)) return -1;
    unlikely(get_footbias(
	       &w2f, &rfoot, imupitch, imuroll, rfootF[2], &rfootbias))
      return -1;
    unlikely(!t3dComp(&w2f, &rfootbias)) return -1;
    unlikely(!t3dCompIn(waist, &w2f, &rfoot)) return -1;

    /* waist from left foot and correct with IMU */
    unlikely(get_waist2foot(angles, HRP2_ODOLEFT, &w2f)) return -1;
    unlikely(get_footbias(
	       &w2f, &lfoot, imupitch, imuroll, lfootF[2], &lfootbias)) return -1;
    unlikely(!t3dComp(&w2f, &lfootbias)) return -1;
    unlikely(!t3dCompIn(&avg, &w2f, &lfoot)) return -1;

    /* average (pitch and roll a computed so that they are equal already */
    unlikely(!t3dConvertTo(T3D_U_BRYAN, waist)) return -1;
    unlikely(!t3dConvertTo(T3D_U_BRYAN, &avg)) return -1;
    waist->ubryan.yaw = (waist->ubryan.yaw + avg.ubryan.yaw) * .5;
    waist->ubryan.x = (waist->ubryan.x + avg.ubryan.x) * .5;
    waist->ubryan.y = (waist->ubryan.y + avg.ubryan.y) * .5;
    waist->ubryan.z = (waist->ubryan.z + avg.ubryan.z) * .5;

  } else if (rfootF[2] > HRP2_FORCETH) {
    /* right foot fixed only */

    /* compute waist from right foot and correct with IMU */
    unlikely(get_waist2foot(angles, HRP2_ODORIGHT, &w2f)) return -1;
    unlikely(get_footbias(
	       &w2f, &rfoot, imupitch, imuroll, rfootF[2], &rfootbias)) return -1;
    unlikely(!t3dComp(&w2f, &rfootbias)) return -1;
    unlikely(!t3dCompIn(waist, &w2f, &rfoot)) return -1;

    /* update left foot position from waist */
    unlikely(get_waist2foot(angles, HRP2_ODOLEFT, &w2f)) return -1;
    unlikely(!t3dCompIn(&lfoot, t3dInvert(&w2f), waist)) return -1;

  }  else if (lfootF[2] > HRP2_FORCETH) {
    /* left foot fixed only */

    /* compute waist from left foot and correct with IMU */
    unlikely(get_waist2foot(angles, HRP2_ODOLEFT, &w2f)) return -1;
    unlikely(
      get_footbias(&w2f, &lfoot, imupitch, imuroll, lfootF[2], &lfootbias)) return -1;
    unlikely(!t3dComp(&w2f, &lfootbias)) return -1;
    unlikely(!t3dCompIn(waist, &w2f, &lfoot)) return -1;

    /* update right foot position from waist */
    unlikely(get_waist2foot(angles, HRP2_ODORIGHT, &w2f)) return -1;
    unlikely(!t3dCompIn(&rfoot, t3dInvert(&w2f), waist)) return -1;
  }

  return 0;
}


/* --- hrp2_odolog --------------------------------------------------- */

/** \brief Log odometry data
 *
 * \param[in] f		Log file descriptor
 *
 * \return 0 on success
 */

int
hrp2_odolog(FILE *f)
{
  T3D p;
  int i;

  /* log right foot position */
  t3dConvertTo(T3D_U_BRYAN, t3dCopyIn(&p, &rfoot));
  fprintf(f,"right-foot: ");
  for(i=0; i<6; i++) fprintf(f, "%12g ", p.ubryan.bryan[i]);
  fprintf(f,"| ");
  for(i=0; i<21; i++) fprintf(f, "%12g ", p.ubryan.v[i]);
  fprintf(f,"\n");

  /* log left foot position */
  t3dConvertTo(T3D_U_BRYAN, t3dCopyIn(&p, &lfoot));
  fprintf(f,"left-foot: ");
  for(i=0; i<6; i++) fprintf(f, "%12g ", p.ubryan.bryan[i]);
  fprintf(f,"| ");
  for(i=0; i<21; i++) fprintf(f, "%12g ", p.ubryan.v[i]);
  fprintf(f,"\n");

  /* log right foot bias */
  t3dConvertTo(T3D_U_BRYAN, t3dCopyIn(&p, &rfootbias));
  fprintf(f,"bias-right-foot: ");
  for(i=0; i<6; i++) fprintf(f, "%12g ", p.ubryan.bryan[i]);
  fprintf(f,"| ");
  for(i=0; i<21; i++) fprintf(f, "%12g ", p.ubryan.v[i]);
  fprintf(f,"\n");

  /* log left foot bias */
  t3dConvertTo(T3D_U_BRYAN, t3dCopyIn(&p, &lfootbias));
  fprintf(f,"bias-left-foot: ");
  for(i=0; i<6; i++) fprintf(f, "%12g ", p.ubryan.bryan[i]);
  fprintf(f,"| ");
  for(i=0; i<21; i++) fprintf(f, "%12g ", p.ubryan.v[i]);
  fprintf(f,"\n");

  return 0;
}


/* --- get_waist2foot ------------------------------------------------ */

/** \brief Compute waist to foot 3d transform.
 */

static int
get_waist2foot(const double angles[GHRP2_MAXDOF], int leg, T3D *rt3d)
{
  T3D t;

  t3dCopyIn(rt3d, &t3dMatrixIdentity);
  rt3d->flags |= T3D_ALLOW_CONVERSION;

  /* hip */
  if (!t3dComp(rt3d, &Thip[leg])) return -1;

  t = t3dUBryanIdentity;
  t.flags |= T3D_ALLOW_CONVERSION;
  t.ubryan.yaw =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT0:GHRP2_LLEG_JOINT0];
  t.ubryan.v[0] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  if (!t3dComp(rt3d, &t)) return -1;

  t = t3dUBryanIdentity;
  t.flags |= T3D_ALLOW_CONVERSION;
  t.ubryan.pitch =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT2:GHRP2_LLEG_JOINT2];
  t.ubryan.v[2] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  t.ubryan.roll =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT1:GHRP2_LLEG_JOINT1];
  t.ubryan.v[5] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  if (!t3dComp(rt3d, &t)) return -1;

  /* thigh */
  if (!t3dComp(rt3d, &Tthigh)) return -1;

  t = t3dUBryanIdentity;
  t.flags |= T3D_ALLOW_CONVERSION;
  t.ubryan.pitch =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT3:GHRP2_LLEG_JOINT3];
  t.ubryan.v[2] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  if (!t3dComp(rt3d, &t)) return -1;

  /* leg */
  if (!t3dComp(rt3d, &Tleg[leg])) return -1;

  t = t3dUBryanIdentity;
  t.flags |= T3D_ALLOW_CONVERSION;
  t.ubryan.pitch =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT4:GHRP2_LLEG_JOINT4];
  t.ubryan.v[2] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  if (!t3dComp(rt3d, &t)) return -1;

  t = t3dUBryanIdentity;
  t.flags |= T3D_ALLOW_CONVERSION;
  t.ubryan.roll =
    -angles[(leg == HRP2_ODORIGHT)?GHRP2_RLEG_JOINT5:GHRP2_LLEG_JOINT5];
  t.ubryan.v[5] = HRP2_ENCSTDDEV*HRP2_ENCSTDDEV;
  if (!t3dComp(rt3d, &t)) return -1;

  return 0;
}


/* --- get_footbias -------------------------------------------------- */

/** \brief Compute roll and pitch bias on the foot
 *
 * By using the roll and pitch angles from the IMU, estimate the unknown
 * roll and pitch noise at the foot level.
 *
 * \param[in] w2f	Waist to foot (biased) tranformation.
 * \param[in] f2W	Foot to world tranformation.
 * \param[in] imupitch	Pitch of the waist from the IMU.
 * \param[in] imuroll	Roll of the waist from the IMU.
 * \param[out] bias	Bias (roll and pitch).
 */

static int
get_footbias(const T3D *w2f, const T3D *f2W,
	     double imupitch, double imuroll, double force,
	     T3D *bias)
{
  T3D odof2W, odorpy, imurpy;

  if (!t3dCopyIn(&odorpy, w2f)) return -1;
  if (!t3dConvertTo(T3D_U_BRYAN, &odorpy)) return -1;
  odorpy.flags = T3D_ALLOW_CONVERSION;
  odorpy.ubryan.yaw = 0.;
  odorpy.ubryan.x = 0.;
  odorpy.ubryan.y = 0.;
  odorpy.ubryan.z = 0.;

  if (!t3dCopyIn(&odof2W, f2W)) return -1;
  if (!t3dConvertTo(T3D_U_BRYAN, &odof2W)) return -1;
  odof2W.flags = T3D_ALLOW_CONVERSION;
  odof2W.ubryan.yaw = 0.;
  odof2W.ubryan.x = 0.;
  odof2W.ubryan.y = 0.;
  odof2W.ubryan.z = 0.;

  if (!t3dCopyIn(&imurpy, &odorpy)) return -1;
  imurpy.flags = T3D_ALLOW_CONVERSION;
  imurpy.ubryan.pitch = imupitch;
  imurpy.ubryan.roll = imuroll;

  if (!t3dCompIn(bias, t3dInvert(&odorpy), &imurpy)) return -1;
  if (!t3dComp(bias, t3dInvert(&odof2W))) return -1;
  if (!t3dConvertTo(T3D_U_BRYAN, bias)) return -1;
  bias->flags = T3D_ALLOW_CONVERSION;
  bias->ubryan.yaw = 0.;
  bias->ubryan.x = 0.;
  bias->ubryan.y = 0.;
  bias->ubryan.z = - force / 90.0e3;

  return 0;
}
